from __future__ import annotations

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import calculate_risk_score, get_logger

from .base import Scanner

LOGGER = get_logger()

# The most performant model. 0.43 B parameters, 870 MB.
# It's English only. Context length max 512 tokens
MODEL_DEBERTA_LARGE_V2 = Model(
    path="MoritzLaurer/deberta-v3-large-zeroshot-v2.0",
    revision="cf44676c28ba7312e5c5f8f8d2c22b3e0c9cdae2",
    onnx_path="MoritzLaurer/deberta-v3-large-zeroshot-v2.0",
    onnx_revision="cf44676c28ba7312e5c5f8f8d2c22b3e0c9cdae2",
    onnx_subfolder="onnx",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
    tokenizer_kwargs={
        "model_input_names": ["input_ids", "attention_mask"],
    },
)

# The most performant base model. 0.18 B parameters, 369 MB.
# It's English only. Context length max 512 tokens.
# Faster than RoBERTa-large/BGE-3 models, but slower than RoBERTa-base
MODEL_DEBERTA_BASE_V2 = Model(
    path="MoritzLaurer/deberta-v3-base-zeroshot-v2.0",
    revision="8e7e5af5983a0ddb1a5b45a38b129ab69e2258e8",
    onnx_path="MoritzLaurer/deberta-v3-base-zeroshot-v2.0",
    onnx_subfolder="onnx",
    onnx_revision="8e7e5af5983a0ddb1a5b45a38b129ab69e2258e8",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
    tokenizer_kwargs={
        "model_input_names": ["input_ids", "attention_mask"],
    },
)

# The most performance multilingual model. 0.57 B parameters, 1.14 GB.
# 100+ languages; context length max 8192 tokens; based on bge-m3-retromae, which is based on XLM-RoBERTa
MODEL_BGE_M3_V2 = Model(
    path="MoritzLaurer/bge-m3-zeroshot-v2.0",
    revision="cd3f8598c7359a3b5cbce164d7fcdafb83a36484",
    onnx_path="MoritzLaurer/bge-m3-zeroshot-v2.0",
    onnx_subfolder="onnx",
    onnx_revision="cd3f8598c7359a3b5cbce164d7fcdafb83a36484",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)

# Less performant than deberta-v3 variants, but a bit faster and compatible with flash attention and TEI containers.
# Size: 0.35B parameters, 711 MB. It's English only. Context length max 512 tokens.
# Only trained on commercially-friendly data.
MODEL_ROBERTA_LARGE_C_V2 = Model(
    path="MoritzLaurer/roberta-large-zeroshot-v2.0-c",
    revision="4c24ed4bba5af8d3162604abc2a141b9d2183ecc",
    onnx_path="MoritzLaurer/roberta-large-zeroshot-v2.0-c",
    onnx_subfolder="onnx",
    onnx_revision="4c24ed4bba5af8d3162604abc2a141b9d2183ecc",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)

# Same model but smaller, more efficient version.
MODEL_ROBERTA_BASE_C_V2 = Model(
    path="MoritzLaurer/roberta-base-zeroshot-v2.0-c",
    revision="d825e740e0c59881cf0b0b1481ccf726b6d65341",
    onnx_path="protectai/MoritzLaurer-roberta-base-zeroshot-v2.0-c-onnx",
    onnx_revision="fde5343dbad32f1a5470890505c72ec656db6dbe",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)


class BanTopics(Scanner):
    """
    BanTopics class is used to ban certain topics from the prompt.

    It uses a HuggingFace model to perform zero-shot classification.
    """

    def __init__(
        self,
        topics: list[str],
        *,
        threshold: float = 0.6,
        model: Model | None = None,
        use_onnx: bool = False,
    ) -> None:
        """
        Initialize BanTopics object.

        Parameters:
            topics (Sequence[str]): List of topics to ban.
            threshold (float, optional): Threshold to determine if a topic is present in the prompt. Default is 0.75.
            model (Model, optional): Model to use for zero-shot classification. Default is roberta-base-c-v2.
            use_onnx (bool, optional): Whether to use ONNX for inference. Default is False.

        Raises:
            ValueError: If no topics are provided.
        """
        self._topics = topics
        self._threshold = threshold

        if model is None:
            model = MODEL_ROBERTA_BASE_C_V2

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._classifier = pipeline(
            task="zero-shot-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

    def scan(self, prompt: str) -> tuple[str, bool, float]:
        if prompt.strip() == "":
            return prompt, True, 0.0

        output_model = self._classifier(prompt, self._topics, multi_label=False)
        label_score = dict(zip(output_model["labels"], output_model["scores"]))

        max_score = round(max(output_model["scores"]) if output_model["scores"] else 0, 2)
        if max_score > self._threshold:
            LOGGER.warning(
                "Topics detected for the prompt",
                scores=label_score,
            )

            return prompt, False, calculate_risk_score(max_score, self._threshold)

        LOGGER.debug(
            "No banned topics detected",
            scores=label_score,
        )

        return prompt, True, 0.0



from __future__ import annotations

import importlib
from functools import lru_cache
from typing import Literal, get_args

from transformers import (
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
    TFPreTrainedModel,
)

from .exception import LLMGuardValidationError
from .model import Model
from .util import device, get_logger, lazy_load_dep

LOGGER = get_logger()


def get_tokenizer(model: Model):
    """
    This function loads a tokenizer given a model identifier and caches it.
    Subsequent calls with the same model_identifier will return the cached tokenizer.

    Args:
        model (Model): The model to load the tokenizer for.
    """
    transformers = lazy_load_dep("transformers")
    tokenizer = transformers.AutoTokenizer.from_pretrained(
        model.path, revision=model.revision, **model.tokenizer_kwargs
    )
    return tokenizer


@lru_cache(maxsize=None)  # Unbounded cache
def is_onnx_supported() -> bool:
    is_supported = importlib.util.find_spec("optimum.onnxruntime") is not None  # type: ignore
    if not is_supported:
        LOGGER.warning(
            "ONNX Runtime is not available. "
            "Please install optimum: "
            "`pip install llm-guard[onnxruntime]` for CPU or "
            "`pip install llm-guard[onnxruntime-gpu]` for GPU to enable ONNX Runtime optimizations."
        )

    return is_supported


def _ort_model_for_sequence_classification(
    model: Model,
):
    provider = "CPUExecutionProvider"
    package_name = "optimum[onnxruntime]"
    if device().type == "cuda":
        package_name = "optimum[onnxruntime-gpu]"
        provider = "CUDAExecutionProvider"

    onnxruntime = lazy_load_dep("optimum.onnxruntime", package_name)

    tf_model = onnxruntime.ORTModelForSequenceClassification.from_pretrained(
        model.onnx_path or model.path,
        export=model.onnx_path is None,
        file_name=model.onnx_filename,
        subfolder=model.onnx_subfolder,
        revision=model.onnx_revision,
        provider=provider,
        **model.kwargs,
    )
    LOGGER.debug("Initialized classification ONNX model", model=model, device=device())

    return tf_model


def get_tokenizer_and_model_for_classification(
    model: Model,
    use_onnx: bool = False,
):
    """
    This function loads a tokenizer and model given a model identifier and caches them.
    Subsequent calls with the same model_identifier will return the cached tokenizer.

    Args:
        model (str): The model identifier to load the tokenizer and model for.
        use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
    """
    tf_tokenizer = get_tokenizer(model)
    transformers = lazy_load_dep("transformers")

    if use_onnx and is_onnx_supported() is False:
        LOGGER.warning("ONNX is not supported on this machine. Using PyTorch instead of ONNX.")
        use_onnx = False

    if use_onnx is False:
        tf_model = transformers.AutoModelForSequenceClassification.from_pretrained(
            model.path, subfolder=model.subfolder, revision=model.revision, **model.kwargs
        )
        LOGGER.debug("Initialized classification model", model=model, device=device())

        return tf_tokenizer, tf_model

    tf_model = _ort_model_for_sequence_classification(model)

    return tf_tokenizer, tf_model


def get_tokenizer_and_model_for_ner(
    model: Model,
    use_onnx: bool = False,
):
    """
    This function loads a tokenizer and model given a model identifier and caches them.
    Subsequent calls with the same model_identifier will return the cached tokenizer.

    Args:
        model (str): The model identifier to load the tokenizer and model for.
        use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
    """
    tf_tokenizer = get_tokenizer(model)
    transformers = lazy_load_dep("transformers")

    if use_onnx and is_onnx_supported() is False:
        LOGGER.warning("ONNX is not supported on this machine. Using PyTorch instead of ONNX.")
        use_onnx = False

    if use_onnx is False:
        tf_model = transformers.AutoModelForTokenClassification.from_pretrained(
            model.path, subfolder=model.subfolder, revision=model.revision, **model.kwargs
        )
        LOGGER.debug("Initialized NER model", model=model, device=device())

        return tf_tokenizer, tf_model

    optimum_onnxruntime = lazy_load_dep(
        "optimum.onnxruntime",
        "optimum[onnxruntime]" if device().type != "cuda" else "optimum[onnxruntime-gpu]",
    )

    tf_model = optimum_onnxruntime.ORTModelForTokenClassification.from_pretrained(
        model.onnx_path,
        export=False,
        subfolder=model.onnx_subfolder,
        provider=("CUDAExecutionProvider" if device().type == "cuda" else "CPUExecutionProvider"),
        revision=model.onnx_revision,
        file_name=model.onnx_filename,
        **model.kwargs,
    )
    LOGGER.debug("Initialized NER ONNX model", model=model, device=device())

    return tf_tokenizer, tf_model


ClassificationTask = Literal["text-classification", "zero-shot-classification"]


def pipeline(
    task: str,
    model: PreTrainedModel | TFPreTrainedModel,
    tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast,
    **kwargs,
):
    if task not in get_args(ClassificationTask):
        raise LLMGuardValidationError(f"Invalid task. Must be one of {ClassificationTask}")

    if kwargs.get("max_length", None) is None:
        kwargs["max_length"] = tokenizer.model_max_length

    transformers = lazy_load_dep("transformers")
    return transformers.pipeline(
        task,
        model=model,
        tokenizer=tokenizer,
        **kwargs,
    )



from __future__ import annotations

import string
from typing import Any, Callable, cast

from faker import Faker

fake = Faker()
fake.seed_instance(100)


_entity_faker_map: dict[str, Callable[[], Any]] = {
    # Global entities
    "CREDIT_CARD": fake.credit_card_number,
    "EMAIL_ADDRESS": fake.email,
    "IBAN_CODE": fake.iban,
    "IP_ADDRESS": fake.ipv4_public,
    "PERSON": fake.name,
    "PHONE_NUMBER": fake.phone_number,
    "URL": fake.url,
    "CREDIT_CARD_RE": fake.credit_card_number,
    "UUID": fake.uuid4,
    "LOCATION": fake.city,
    "DATE_TIME": fake.date,
    "CRYPTO": cast(
        Callable[[], str],
        lambda _: "bc1"
        + "".join(fake.random_choices(string.ascii_lowercase + string.digits, length=26)),
    ),
    "NRP": cast(Callable[[], str], lambda _: str(fake.random_number(digits=8, fix_len=True))),
    "MEDICAL_LICENSE": cast(Callable[[], str], lambda _: fake.bothify(text="??######").upper()),
    # US-specific entities
    "US_BANK_NUMBER": fake.bban,
    "US_SSN": fake.ssn,
    "US_DRIVER_LICENSE": cast(
        Callable[[], str], lambda _: str(fake.random_number(digits=9, fix_len=True))
    ),
    "US_ITIN": cast(Callable[[], str], lambda _: fake.bothify(text="9##-7#-####")),
    "US_PASSPORT": cast(Callable[[], str], lambda _: fake.bothify(text="#####??").upper()),
    # UK-specific entities
    "UK_NHS": cast(Callable[[], str], lambda _: str(fake.random_number(digits=10, fix_len=True))),
    # Spain-specific entities
    "ES_NIF": cast(Callable[[], str], lambda _: fake.bothify(text="########?").upper()),
    # Italy-specific entities
    "IT_FISCAL_CODE": cast(
        Callable[[], str], lambda _: fake.bothify(text="??????##?##?###?").upper()
    ),
    "IT_DRIVER_LICENSE": cast(Callable[[], str], lambda _: fake.bothify(text="?A#######?").upper()),
    "IT_VAT_CODE": cast(Callable[[], str], lambda _: fake.bothify(text="IT???????????")),
    "IT_PASSPORT": cast(
        Callable[[], str], lambda _: str(fake.random_number(digits=9, fix_len=True))
    ),
    "IT_IDENTITY_CARD": cast(
        Callable[[], str], lambda _: lambda _: str(fake.random_number(digits=7, fix_len=True))
    ),
    # Singapore-specific entities
    "SG_NRIC_FIN": cast(Callable[[], str], lambda _: fake.bothify(text="????####?").upper()),
    # Australia-specific entities
    "AU_ABN": cast(Callable[[], str], lambda _: str(fake.random_number(digits=11, fix_len=True))),
    "AU_ACN": cast(Callable[[], str], lambda _: str(fake.random_number(digits=9, fix_len=True))),
    "AU_TFN": cast(Callable[[], str], lambda _: str(fake.random_number(digits=9, fix_len=True))),
    "AU_MEDICARE": cast(
        Callable[[], str], lambda _: str(fake.random_number(digits=10, fix_len=True))
    ),
}


def get_fake_value(entity_type: str) -> str | None:
    if entity_type not in _entity_faker_map:
        return None

    return _entity_faker_map[entity_type]()



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence, cast

from presidio_anonymizer.core.text_replace_builder import TextReplaceBuilder

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_ner
from llm_guard.util import get_logger, lazy_load_dep

from .base import Scanner

if TYPE_CHECKING:
    import transformers

LOGGER = get_logger()

MODEL_V1 = Model(
    path="guishe/nuner-v1_orgs",
    revision="2e95454e741e5bdcbfabd6eaed5fb03a266cf043",
    onnx_path="protectai/guishe-nuner-v1_orgs-onnx",
    onnx_revision="20c9739f45f6b4d10ba63c62e6fa92f214a12a52",
    onnx_subfolder="",
    pipeline_kwargs={
        "aggregation_strategy": "simple",
    },
)


class BanCompetitors(Scanner):
    """
    Scanner to detect competitors in a prompt.

    It uses a named-entity recognition model to extract organizations from the prompt and compares them to a list of competitors.
    """

    def __init__(
        self,
        competitors: Sequence[str],
        *,
        threshold: float = 0.5,
        redact: bool = True,
        model: Model | None = None,
        use_onnx: bool = False,
    ) -> None:
        """
        Initialize BanCompetitors object.

        Parameters:
            competitors (Sequence[str]): List of competitors to detect.
            threshold (float, optional): Threshold to determine if a competitor is present in the prompt. Default is 0.5.
            redact (bool, optional): Whether to redact the competitor name. Default is True.
            model (Model, optional): Model to use for named-entity recognition. Default is V1 model.
            use_onnx (bool, optional): Whether to use ONNX instead of PyTorch for inference. Default is False.

        Raises:
            ValueError: If no topics are provided.
        """
        if model is None:
            model = MODEL_V1

        self._competitors = competitors
        self._threshold = threshold
        self._redact = redact

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_ner(
            model=model,
            use_onnx=use_onnx,
        )

        transformers = cast("transformers", lazy_load_dep("transformers"))
        self._ner_pipeline = transformers.pipeline(
            "ner", model=tf_model, tokenizer=tf_tokenizer, **model.pipeline_kwargs
        )

    def scan(self, prompt: str) -> tuple[str, bool, float]:
        is_detected = False
        text_replace_builder = TextReplaceBuilder(original_text=prompt)
        entities = self._ner_pipeline(prompt)
        assert isinstance(entities, list)
        entities = sorted(entities, key=lambda x: x["end"], reverse=True)

        for entity in entities:
            entity["word"] = entity["word"].strip()
            if entity["word"] not in self._competitors:
                LOGGER.debug("Entity is not a specified competitor", entity=entity["word"])
                continue

            if entity["score"] < self._threshold:
                LOGGER.debug(
                    "Competitor detected but the score is below threshold",
                    entity=entity["word"],
                    score=entity["score"],
                )
                continue

            is_detected = True

            if self._redact:
                text_replace_builder.replace_text_get_insertion_index(
                    "[REDACTED]",
                    entity["start"],
                    entity["end"],
                )

            LOGGER.warning(
                "Competitor detected with score", entity=entity["word"], score=entity["score"]
            )

        if is_detected:
            return text_replace_builder.output_text, False, 1.0

        LOGGER.debug("None of the competitors were detected")

        return prompt, True, 0.0



from __future__ import annotations

import os
import re
from typing import Final

from presidio_analyzer import RecognizerResult
from presidio_anonymizer.core.text_replace_builder import TextReplaceBuilder

from llm_guard.input_scanners.anonymize_helpers.ner_mapping import NERConfig

from ..exception import LLMGuardValidationError
from ..util import calculate_risk_score, get_logger
from ..vault import Vault
from .anonymize_helpers import (
    DEBERTA_AI4PRIVACY_v2_CONF,
    get_analyzer,
    get_fake_value,
    get_regex_patterns,
    get_transformers_recognizer,
)
from .anonymize_helpers.regex_patterns import DefaultRegexPatterns, RegexPatternsReuse
from .base import Scanner

LOGGER = get_logger()

DEFAULT_ENTITY_TYPES: Final[list[str]] = [
    "CREDIT_CARD",
    "CRYPTO",
    "EMAIL_ADDRESS",
    "IBAN_CODE",
    "IP_ADDRESS",
    "PERSON",
    "PHONE_NUMBER",
    "US_SSN",
    "US_BANK_NUMBER",
    "CREDIT_CARD_RE",
    "UUID",
    "EMAIL_ADDRESS_RE",
    "US_SSN_RE",
]

ALL_SUPPORTED_LANGUAGES: Final[list[str]] = ["en", "zh"]


class Anonymize(Scanner):
    """
    Anonymize sensitive data in the text using NLP (English only) and predefined regex patterns.

    Anonymizes detected entities with placeholders like [REDACTED_PERSON_1] and stores the real values in a Vault.
    Deanonymizer can be used to replace the placeholders back to their original values.
    """

    def __init__(
        self,
        vault: Vault,
        *,
        hidden_names: list[str] | None = None,
        allowed_names: list[str] | None = None,
        entity_types: list[str] | None = None,
        preamble: str = "",
        regex_patterns: list[DefaultRegexPatterns | RegexPatternsReuse] | None = None,
        use_faker: bool = False,
        recognizer_conf: NERConfig | None = None,
        threshold: float = 0.5,
        use_onnx: bool = False,
        language: str = "en",
    ) -> None:
        """
        Initialize an instance of Anonymize class.

        Parameters:
            vault: A vault instance to store the anonymized data.
            hidden_names: List of names to be anonymized e.g. [REDACTED_CUSTOM_1].
            allowed_names: List of names allowed in the text without anonymizing.
            entity_types: List of entity types to be detected. If not provided, defaults to all.
            preamble: Text to prepend to sanitized prompt. If not provided, defaults to an empty string.
            regex_patterns: List of regex patterns to be used for detection. If not provided, defaults to predefined list.
            use_faker: Whether to use faker instead of placeholders in applicable cases. If not provided, defaults to False, replaces with placeholders [REDACTED_PERSON_1].
            recognizer_conf: Configuration to recognize PII data. Default is Ai4Privacy DeBERTa model.
            threshold: Acceptance threshold. Default is 0.
            use_onnx: Whether to use ONNX runtime for inference. Default is False.
            language: Language of the anonymize detect. Default is "en".
        """

        if language not in ALL_SUPPORTED_LANGUAGES:
            raise LLMGuardValidationError(
                f"Language must be in the list of allowed: {ALL_SUPPORTED_LANGUAGES}"
            )

        os.environ["TOKENIZERS_PARALLELISM"] = "false"  # Disables huggingface/tokenizers warning

        if not entity_types:
            LOGGER.debug(
                "No entity types provided, using default", default_entities=DEFAULT_ENTITY_TYPES
            )
            entity_types = DEFAULT_ENTITY_TYPES.copy()

        entity_types.append("CUSTOM")

        if not hidden_names:
            hidden_names = []

        self._vault = vault
        self._entity_types = entity_types
        self._allowed_names = allowed_names
        self._preamble = preamble
        self._use_faker = use_faker
        self._threshold = threshold
        self._language = language

        if not recognizer_conf:
            recognizer_conf = DEBERTA_AI4PRIVACY_v2_CONF

        transformers_recognizer = get_transformers_recognizer(
            recognizer_conf=recognizer_conf,
            use_onnx=use_onnx,
            supported_language=language,
        )

        self._analyzer = get_analyzer(
            recognizer=transformers_recognizer,
            regex_groups=get_regex_patterns(regex_patterns),
            custom_names=hidden_names,
            supported_languages=ALL_SUPPORTED_LANGUAGES,
        )

    def _remove_conflicts_and_get_text_manipulation_data(
        self, analyzer_results: list[RecognizerResult]
    ) -> list[RecognizerResult]:
        """
        Iterate the list and create a sorted unique results list from it.

        Only insert results which are:
        1. Indices are not contained in other result.
        2. Have the same indices as other results but with larger score.
        :return: list
        """
        tmp_analyzer_results = []
        # This list contains all elements which we need to check a single result
        # against. If a result is dropped, it can also be dropped from this list
        # since it is intersecting with another result and we selected the other one.
        other_elements = analyzer_results.copy()
        for result in analyzer_results:
            other_elements.remove(result)

            is_merge_same_entity_type = False
            for other_element in other_elements:
                if other_element.entity_type != result.entity_type:
                    continue
                if result.intersects(other_element) == 0:
                    continue

                other_element.start = min(result.start, other_element.start)
                other_element.end = max(result.end, other_element.end)
                other_element.score = max(result.score, other_element.score)
                is_merge_same_entity_type = True
                break
            if not is_merge_same_entity_type:
                other_elements.append(result)
                tmp_analyzer_results.append(result)
            else:
                LOGGER.debug(f"removing element {result} from " f"results list due to merge")

        unique_text_metadata_elements = []
        # This list contains all elements which we need to check a single result
        # against. If a result is dropped, it can also be dropped from this list
        # since it is intersecting with another result and we selected the other one.
        other_elements = tmp_analyzer_results.copy()
        for result in tmp_analyzer_results:
            other_elements.remove(result)
            result_conflicted = self.__is_result_conflicted_with_other_elements(
                other_elements, result
            )
            if not result_conflicted:
                other_elements.append(result)
                unique_text_metadata_elements.append(result)
            else:
                LOGGER.debug(f"removing element {result} from results list due to conflict")

        # This further improves the quality of handling the conflict between the
        # various entities overlapping. This will not drop the results instead
        # it adjust the start and end positions of overlapping results and removes
        # All types of conflicts among entities as well as text.
        unique_text_metadata_elements.sort(key=lambda element: element.start)
        elements_length = len(unique_text_metadata_elements)
        index = 0
        while index < elements_length - 1:
            current_entity = unique_text_metadata_elements[index]
            next_entity = unique_text_metadata_elements[index + 1]
            if current_entity.end <= next_entity.start:
                index += 1
            else:
                if current_entity.score >= next_entity.score:
                    next_entity.start = current_entity.end
                else:
                    current_entity.end = next_entity.start
                unique_text_metadata_elements.sort(key=lambda element: element.start)
        unique_text_metadata_elements = [
            element for element in unique_text_metadata_elements if element.start <= element.end
        ]
        return unique_text_metadata_elements

    @staticmethod
    def __is_result_conflicted_with_other_elements(other_elements, result):
        """
        Check if the given result conflicts with any other elements.
        """
        return any([result.has_conflict(other_element) for other_element in other_elements])

    @staticmethod
    def _merge_entities_with_whitespace_between(
        text: str, analyzer_results: list[RecognizerResult]
    ) -> list[RecognizerResult]:
        """
        Merge adjacent entities of the same type separated by whitespace.
        """
        merged_results = []
        prev_result = None
        for result in analyzer_results:
            if prev_result is not None:
                if prev_result.entity_type == result.entity_type:
                    if re.search(r"^( )+$", text[prev_result.end : result.start]):
                        merged_results.remove(prev_result)
                        result.start = prev_result.start
            merged_results.append(result)
            prev_result = result
        return merged_results

    @staticmethod
    def _get_entity_placeholder(entity_type: str, index: int, use_faker: bool) -> str:
        result = f"[REDACTED_{entity_type}_{index}]"
        if use_faker:
            result = get_fake_value(entity_type) or result

        return result

    @staticmethod
    def _anonymize(
        prompt: str, pii_entities: list[RecognizerResult], vault: Vault, use_faker: bool
    ) -> tuple[str, list[tuple[str, str]]]:
        """
        Replace detected entities in the prompt with anonymized placeholders.

        Parameters:
            prompt: Original text prompt.
            pii_entities: List of entities detected in the prompt.
            vault: A vault instance with the anonymized data stored.
            use_faker: Whether to use faker to generate fake data.

        Returns:
            str: Sanitized text.
            list[tuple]: list of tuples representing the replaced entities and their corresponding placeholders.
        """
        text_replace_builder = TextReplaceBuilder(original_text=prompt)

        entity_type_counter, new_entity_counter = {}, {}
        for pii_entity in pii_entities:
            entity_type = pii_entity.entity_type
            entity_value = text_replace_builder.get_text_in_position(
                pii_entity.start, pii_entity.end
            )

            if entity_type not in entity_type_counter:
                entity_type_counter[entity_type] = {}

            if entity_value not in entity_type_counter[entity_type]:
                vault_entities = [
                    (entity_placeholder, entity_vault_value)
                    for entity_placeholder, entity_vault_value in vault.get()
                    if entity_type in entity_placeholder
                ]
                entity_placeholder = [
                    entity_placeholder
                    for entity_placeholder, entity_vault_value in vault_entities
                    if entity_vault_value == entity_value
                ]
                if len(entity_placeholder) > 0:
                    entity_type_counter[entity_type][entity_value] = int(
                        entity_placeholder[0].split("_")[-1][:-1]
                    )
                else:
                    entity_type_counter[entity_type][entity_value] = (
                        len(vault_entities) + new_entity_counter.get(entity_type, 0) + 1
                    )
                    new_entity_counter[entity_type] = new_entity_counter.get(entity_type, 0) + 1

        results = []
        sorted_pii_entities = sorted(pii_entities, reverse=True)
        for pii_entity in sorted_pii_entities:
            entity_type = pii_entity.entity_type
            entity_value = text_replace_builder.get_text_in_position(
                pii_entity.start, pii_entity.end
            )

            index = entity_type_counter[entity_type][entity_value]
            changed_entity = Anonymize._get_entity_placeholder(entity_type, index, use_faker)

            results.append((changed_entity, prompt[pii_entity.start : pii_entity.end]))

            text_replace_builder.replace_text_get_insertion_index(
                changed_entity, pii_entity.start, pii_entity.end
            )

        return text_replace_builder.output_text, results

    @staticmethod
    def remove_single_quotes(text: str) -> str:
        text_without_single_quotes = text.replace("'", " ")
        return text_without_single_quotes

    def scan(self, prompt: str) -> tuple[str, bool, float]:
        risk_score = 0.0
        if prompt.strip() == "":
            return prompt, True, risk_score

        analyzer_results = self._analyzer.analyze(
            text=Anonymize.remove_single_quotes(prompt),
            language=self._language,
            entities=self._entity_types,
            allow_list=self._allowed_names,
            score_threshold=self._threshold,
        )

        risk_score = round(
            (
                max(analyzer_result.score for analyzer_result in analyzer_results)
                if analyzer_results
                else 0.0
            ),
            2,
        )
        analyzer_results = self._remove_conflicts_and_get_text_manipulation_data(analyzer_results)
        merged_results = self._merge_entities_with_whitespace_between(prompt, analyzer_results)

        sanitized_prompt, anonymized_results = self._anonymize(
            prompt, merged_results, self._vault, self._use_faker
        )

        if prompt != sanitized_prompt:
            LOGGER.warning(
                "Found sensitive data in the prompt and replaced it",
                merged_results=merged_results,
                risk_score=risk_score,
            )
            for entity_placeholder, entity_value in anonymized_results:
                if not self._vault.placeholder_exists(entity_placeholder):
                    self._vault.append((entity_placeholder, entity_value))
            return (
                self._preamble + sanitized_prompt,
                False,
                calculate_risk_score(risk_score, self._threshold),
            )

        LOGGER.debug("Prompt does not have sensitive data to replace", risk_score=risk_score)

        return prompt, True, 0.0



from __future__ import annotations

import re

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import calculate_risk_score, get_logger, remove_markdown

from .base import Scanner

LOGGER = get_logger()

MODEL_SM = Model(
    path="vishnun/codenlbert-sm",
    revision="caa3d167fd262c76c7da23cd72c1d24cfdcafd0f",
    onnx_path="protectai/vishnun-codenlbert-sm-onnx",
    onnx_revision="2b1d298410bd98832e41e3da82e20f6d8dff1bc7",
    pipeline_kwargs={"max_length": 128, "truncation": True, "return_token_type_ids": True},
)

MODEL_TINY = Model(
    path="vishnun/codenlbert-tiny",
    revision="2caf5a621b29c50038ee081479a82f192e9a5e69",
    onnx_path="protectai/vishnun-codenlbert-tiny-onnx",
    onnx_revision="84148cb4b3f08fe44705e2d8ed81505450ae8abd",
    pipeline_kwargs={"max_length": 128, "truncation": True, "return_token_type_ids": True},
)


class BanCode(Scanner):
    """
    A scanner that detects if input is code and blocks it.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.97,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes the BanCode scanner.

        Parameters:
           model (Model, optional): The model object.
           threshold (float): The probability threshold. Default is 0.97.
           use_onnx (bool): Whether to use ONNX instead of PyTorch for inference.
        """

        self._threshold = threshold
        if model is None:
            model = MODEL_SM

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._classifier = pipeline(
            task="text-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

    def scan(self, prompt: str) -> tuple[str, bool, float]:
        if prompt.strip() == "":
            return prompt, True, 0.0

        # Hack: Improve accuracy
        new_prompt = remove_markdown(prompt)  # Remove markdown
        new_prompt = re.sub(r"\d+\.\s+|[-*•]\s+", "", new_prompt)  # Remove list markers
        new_prompt = re.sub(r"\d+", "", new_prompt)  # Remove numbers
        new_prompt = re.sub(r'\.(?!\d)(?=[\s\'"“”‘’)\]}]|$)', "", new_prompt)  # Remove periods

        result = self._classifier(new_prompt)[0]
        score = round(
            result["score"] if result["label"] in "CODE" else 1 - result["score"],
            2,
        )

        if score > self._threshold:
            LOGGER.warning(
                "Detected code in the text", score=score, threshold=self._threshold, text=new_prompt
            )

            return prompt, False, calculate_risk_score(score, self._threshold)

        LOGGER.debug(
            "No code detected in the text", score=score, threshold=self._threshold, text=new_prompt
        )

        return prompt, True, 0.0



from __future__ import annotations

from presidio_anonymizer import AnonymizerEngine

from llm_guard.input_scanners.anonymize import (
    ALL_SUPPORTED_LANGUAGES,
    DEFAULT_ENTITY_TYPES,
    Anonymize,
)
from llm_guard.input_scanners.anonymize_helpers import (
    DEBERTA_AI4PRIVACY_v2_CONF,
    get_analyzer,
    get_regex_patterns,
    get_transformers_recognizer,
)
from llm_guard.input_scanners.anonymize_helpers.ner_mapping import NERConfig
from llm_guard.util import calculate_risk_score, get_logger

from ..input_scanners.anonymize_helpers.regex_patterns import (
    DefaultRegexPatterns,
    RegexPatternsReuse,
)
from .base import Scanner

LOGGER = get_logger()


class Sensitive(Scanner):
    """
    A class used to detect sensitive (PII) data in the output of a language model.

    This class uses the Presidio Analyzer Engine and predefined internally patterns (patterns.py) to analyze the output for specified entity types.
    If no entity types are specified, it defaults to checking for all entity types.
    """

    def __init__(
        self,
        *,
        entity_types: list[str] | None = None,
        regex_patterns: list[DefaultRegexPatterns | RegexPatternsReuse] | None = None,
        redact: bool = False,
        recognizer_conf: NERConfig | None = None,
        threshold: float = 0.5,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes an instance of the Sensitive class.

        Parameters:
           entity_types (Optional[Sequence[str]]): The entity types to look for in the output. Defaults to all
                                               entity types.
           regex_patterns (Optional[List[Dict]]): List of regex patterns to use for detection. Default is None.
           redact (bool): Redact found sensitive entities. Default to False.
           recognizer_conf (Optional[Dict]): Configuration to recognize PII data. Default is Ai4Privacy DeBERTa.
           threshold (float): Acceptance threshold. Default is 0.
           use_onnx (bool): Use ONNX model for inference. Default is False.
        """
        if not entity_types:
            LOGGER.debug(
                "No entity types provided, using default", default_entity_types=DEFAULT_ENTITY_TYPES
            )
            entity_types = DEFAULT_ENTITY_TYPES.copy()
        entity_types.append("CUSTOM")

        self._entity_types = entity_types
        self._redact = redact
        self._threshold = threshold

        if not recognizer_conf:
            recognizer_conf = DEBERTA_AI4PRIVACY_v2_CONF

        transformers_recognizer = get_transformers_recognizer(
            recognizer_conf=recognizer_conf,
            use_onnx=use_onnx,
        )
        self._analyzer = get_analyzer(
            transformers_recognizer, get_regex_patterns(regex_patterns), [], ALL_SUPPORTED_LANGUAGES
        )
        self._anonymizer = AnonymizerEngine()

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        if output.strip() == "":
            return prompt, True, 0.0

        analyzer_results = self._analyzer.analyze(
            text=Anonymize.remove_single_quotes(output),
            language="en",
            entities=self._entity_types,
            score_threshold=self._threshold,
        )

        if analyzer_results:
            if self._redact:
                LOGGER.debug("Redacting sensitive entities")
                result = self._anonymizer.anonymize(text=output, analyzer_results=analyzer_results)  # type: ignore
                output = result.text

            risk_score = round(
                max(analyzer_result.score for analyzer_result in analyzer_results), 2
            )
            LOGGER.warning("Found sensitive data in the output", results=analyzer_results)
            return output, False, calculate_risk_score(risk_score, self._threshold)

        LOGGER.debug("No sensitive data found in the output")
        return output, True, 0.0



from __future__ import annotations

from enum import Enum

from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import (
    calculate_risk_score,
    get_logger,
    split_text_by_sentences,
    split_text_to_word_chunks,
    truncate_tokens_head_tail,
)

from .base import Scanner

LOGGER = get_logger()

PROMPT_CHARACTERS_LIMIT = 256

# This model is proprietary but open source.
V1_MODEL = Model(
    path="protectai/deberta-v3-base-prompt-injection",
    revision="f51c3b2a5216ae1af467b511bc7e3b78dc4a99c9",
    onnx_path="ProtectAI/deberta-v3-base-prompt-injection",
    onnx_revision="f51c3b2a5216ae1af467b511bc7e3b78dc4a99c9",
    onnx_subfolder="onnx",
    onnx_filename="model.onnx",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)

V2_MODEL = Model(
    path="protectai/deberta-v3-base-prompt-injection-v2",
    revision="89b085cd330414d3e7d9dd787870f315957e1e9f",
    onnx_path="ProtectAI/deberta-v3-base-prompt-injection-v2",
    onnx_revision="89b085cd330414d3e7d9dd787870f315957e1e9f",
    onnx_subfolder="onnx",
    onnx_filename="model.onnx",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)

# This is gated model, which requires our approval.
V2_SMALL_MODEL = Model(
    path="protectai/deberta-v3-small-prompt-injection-v2",
    revision="3897fe66d47d2e1649e669c3470bf2ba3ddecc22",
    onnx_path="protectai/deberta-v3-small-prompt-injection-v2",
    onnx_revision="3897fe66d47d2e1649e669c3470bf2ba3ddecc22",
    onnx_subfolder="onnx",
    onnx_filename="model.onnx",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
    tokenizer_kwargs={
        "use_fast": False,
        "token": True,
    },
    kwargs={"token": True},  # You can also configure with your token.
)


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"
    # TRUNCATE_TOKEN_HEAD_TAIL is used to split the prompt into two parts (126 head and 382 tail) and check them.
    TRUNCATE_TOKEN_HEAD_TAIL = "truncate_token_head_tail"
    # TRUNCATE_SIDES is used to split the prompt into two parts (256 head and 256 tail) and check them.
    TRUNCATE_HEAD_TAIL = "truncate_head_tail"
    CHUNKS = "chunks"

    _tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast

    def set_tokenizer(self, tokenizer):
        self._tokenizer = tokenizer

    def get_inputs(self, prompt: str) -> list[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        if self == MatchType.CHUNKS:
            chunks = []
            for chunk_start, chunk_end in split_text_to_word_chunks(
                len(prompt), chunk_length=PROMPT_CHARACTERS_LIMIT, overlap_length=25
            ):
                chunks.append(prompt[chunk_start:chunk_end])

            return chunks

        if self == MatchType.TRUNCATE_TOKEN_HEAD_TAIL and self._tokenizer is not None:
            tokenized_input = self._tokenizer.tokenize(prompt)

            return [
                self._tokenizer.convert_tokens_to_string(truncate_tokens_head_tail(tokenized_input))
            ]

        if self == MatchType.TRUNCATE_HEAD_TAIL and len(prompt) > PROMPT_CHARACTERS_LIMIT:
            part_length = (PROMPT_CHARACTERS_LIMIT - 3) // 2

            start = prompt[:part_length]
            end = prompt[-part_length:]

            return [f"{start}...{end}"]

        return [prompt]


class PromptInjection(Scanner):
    """
    A prompt injection scanner based on HuggingFace model. It is used to
    detect if a prompt is attempting to perform an injection attack.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.92,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes PromptInjection with a threshold.

        Parameters:
            model (Model, optional): Chosen model to classify prompt. Default is Laiyer's one.
            threshold (float): Threshold for the injection score. Default is 0.9.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Defaults to False.

        Raises:
            ValueError: If non-existent models were provided.
        """
        if model is None:
            model = V2_MODEL

        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._model = model

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._pipeline = pipeline(
            task="text-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

        match_type.set_tokenizer(tf_tokenizer)
        self._match_type = match_type

    def scan(self, prompt: str) -> tuple[str, bool, float]:
        if prompt.strip() == "":
            return prompt, True, 0.0

        highest_score = 0.0
        results_all = self._pipeline(self._match_type.get_inputs(prompt))
        for result in results_all:
            injection_score = round(
                result["score"] if result["label"] == "INJECTION" else 1 - result["score"],
                2,
            )

            if injection_score > highest_score:
                highest_score = injection_score

            if injection_score > self._threshold:
                LOGGER.warning("Detected prompt injection", injection_score=injection_score)

                return prompt, False, calculate_risk_score(injection_score, self._threshold)

        LOGGER.debug("No prompt injection detected", highest_score=highest_score)

        return prompt, True, 0.0



from __future__ import annotations

import dataclasses

from .util import device


@dataclasses.dataclass
class Model:
    """
    Dataclass to store model information.

    Attributes:
        path (str): Path to the model.
        subfolder (str): Subfolder in the model path.
        revision (str, optional): Revision of the model.
        onnx_path (str, optional): Path to the ONNX model.
        onnx_revision (str, optional): Revision of the ONNX model.
        onnx_subfolder (str): Subfolder in the ONNX model path.
        kwargs (Dict, optional): Keyword arguments passed to the model (transformers).
        pipeline_kwargs (Dict, optional): Keyword arguments passed to the pipeline (transformers).
    """

    path: str
    subfolder: str = ""
    revision: str | None = None
    onnx_path: str | None = None
    onnx_revision: str | None = None
    onnx_subfolder: str = ""
    onnx_filename: str = "model.onnx"
    kwargs: dict = dataclasses.field(default_factory=dict)
    pipeline_kwargs: dict = dataclasses.field(default_factory=dict)
    tokenizer_kwargs: dict = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        default_pipeline_kwargs = {
            "batch_size": 1,
            "device": device(),
        }
        self.pipeline_kwargs = {**default_pipeline_kwargs, **self.pipeline_kwargs}

    def __str__(self):
        return self.path



from __future__ import annotations

from llm_guard.input_scanners.language import Language as InputLanguage, MatchType
from llm_guard.model import Model

from .base import Scanner


class Language(Scanner):
    """
    Language scanner is responsible for determining the language of a given text
    prompt and verifying its validity against a list of predefined languages.
    """

    def __init__(
        self,
        valid_languages: list[str],
        *,
        model: Model | None = None,
        threshold: float = 0.7,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes the Language scanner with a list of valid languages.

        Parameters:
            model (Model, optional): A Model object containing the path to the model and its ONNX equivalent.
            valid_languages (Sequence[str]): A list of valid language codes.
            threshold (float): Minimum confidence score.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
        """

        self._scanner = InputLanguage(
            valid_languages,
            model=model,
            threshold=threshold,
            match_type=match_type,
            use_onnx=use_onnx,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        return self._scanner.scan(output)



from __future__ import annotations

from llm_guard.input_scanners.ban_code import BanCode as InputBanCode
from llm_guard.model import Model

from .base import Scanner


class BanCode(Scanner):
    """
    A scanner that detects code snippets in the model output and blocks them.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.9,
        use_onnx: bool = False,
    ) -> None:
        """
        Initialize a new BanCode scanner.

        Parameters:
            model: The model to use for scanning.
            threshold: The threshold for the model output to consider a code snippet.
            use_onnx: Whether to use the ONNX model for scanning.
        """

        self._scanner = InputBanCode(
            model=model,
            threshold=threshold,
            use_onnx=use_onnx,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        return self._scanner.scan(output)



from __future__ import annotations

from llm_guard.input_scanners.toxicity import MatchType, Toxicity as InputToxicity
from llm_guard.model import Model

from .base import Scanner


class Toxicity(Scanner):
    """
    A class used to detect toxicity in the output of a language model.

    This class uses a pre-trained toxicity model from HuggingFace to calculate a toxicity score for the output.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.7,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes an instance of the Toxicity class.

        Parameters:
            model: The path to the model. Defaults to None.
            threshold: The threshold used to determine toxicity. Defaults to 0.7.
            match_type: Whether to match the full text or individual sentences. Defaults to MatchType.FULL.
            use_onnx: Whether to use ONNX for inference. Defaults to False.
        """

        self._scanner = InputToxicity(
            model=model,
            threshold=threshold,
            match_type=match_type,
            use_onnx=use_onnx,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        return self._scanner.scan(output)



import copy

import spacy
from presidio_analyzer import (
    AnalyzerEngine,
    EntityRecognizer,
    Pattern,
    PatternRecognizer,
    RecognizerRegistry,
)
from presidio_analyzer.context_aware_enhancers import LemmaContextAwareEnhancer
from presidio_analyzer.nlp_engine import NlpEngine, NlpEngineProvider
from spacy.cli import download  # type: ignore

from .ner_mapping import NERConfig
from .predefined_recognizers import _get_predefined_recognizers
from .predefined_recognizers.zh import CustomPatternRecognizer
from .regex_patterns import RegexPattern
from .transformers_recognizer import TransformersRecognizer


def _add_recognizers(
    registry: RecognizerRegistry,
    regex_groups: list[RegexPattern],
    custom_names: list[str],
    supported_languages: list[str] = ["en"],
) -> RecognizerRegistry:
    """
    Create a RecognizerRegistry and populate it with regex patterns and custom names.

    Parameters:
        regex_groups: List of regex patterns.
        custom_names: List of custom names to recognize.

    Returns:
        RecognizerRegistry: A RecognizerRegistry object loaded with regex and custom name recognizers.
    """

    for language in supported_languages:
        # custom recognizer per language
        if len(custom_names) > 0:
            custom_recognizer = PatternRecognizer

            if language == "zh":
                custom_recognizer = CustomPatternRecognizer

            registry.add_recognizer(
                custom_recognizer(
                    supported_entity="CUSTOM",
                    supported_language=language,
                    deny_list=custom_names,
                )
            )

        # predefined recognizers per language
        for _Recognizer in _get_predefined_recognizers(language):
            registry.add_recognizer(_Recognizer(supported_language=language))

    for pattern_data in regex_groups:
        languages = pattern_data["languages"] or ["en"]
        label = pattern_data["name"]
        reuse = pattern_data.get("reuse", False)

        patterns: list[Pattern] = list(
            map(
                lambda exp: Pattern(name=label, regex=exp, score=pattern_data["score"]),
                pattern_data.get("expressions", []) or [],
            )
        )

        for language in languages:
            if language not in supported_languages:
                continue

            if isinstance(reuse, dict):
                new_recognizer = copy.deepcopy(
                    registry.get_recognizers(language=reuse["language"], entities=[reuse["name"]])[
                        0
                    ]
                )
                new_recognizer.supported_language = language
                registry.add_recognizer(new_recognizer)
            else:
                registry.add_recognizer(
                    PatternRecognizer(
                        supported_entity=label,
                        supported_language=language,
                        patterns=patterns,
                        context=pattern_data["context"],
                    )
                )

    return registry


def _get_nlp_engine(languages: list[str] = ["en"]) -> NlpEngine:
    models = []

    for language in languages:
        if not spacy.util.is_package(f"{language}_core_web_sm"):
            # Use small spacy model, for faster inference.
            download(f"{language}_core_web_sm")
        models.append({"lang_code": language, "model_name": f"{language}_core_web_sm"})

    configuration = {"nlp_engine_name": "spacy", "models": models}

    return NlpEngineProvider(nlp_configuration=configuration).create_engine()


def get_transformers_recognizer(
    *,
    recognizer_conf: NERConfig,
    use_onnx: bool = False,
    supported_language: str = "en",
) -> EntityRecognizer:
    """
    This function loads a transformers recognizer given a recognizer configuration.

    Args:
        recognizer_conf: Configuration to recognize PII data.
        use_onnx: Whether to use the ONNX version of the model. Default is False.
        supported_language: The language to use for the recognizer. Default is "en".
    """
    model = recognizer_conf.get("DEFAULT_MODEL")
    supported_entities = recognizer_conf.get("PRESIDIO_SUPPORTED_ENTITIES")
    transformers_recognizer = TransformersRecognizer(
        model=model,
        supported_entities=supported_entities,
        supported_language=supported_language,
    )
    transformers_recognizer.load_transformer(
        use_onnx=use_onnx,
        **recognizer_conf,
    )
    return transformers_recognizer


def get_analyzer(
    recognizer: EntityRecognizer,
    regex_groups: list[RegexPattern],
    custom_names: list[str],
    supported_languages: list[str],
) -> AnalyzerEngine:
    nlp_engine = _get_nlp_engine(languages=supported_languages)

    registry = RecognizerRegistry()
    registry.load_predefined_recognizers(nlp_engine=nlp_engine)
    registry = _add_recognizers(registry, regex_groups, custom_names, supported_languages)
    registry.add_recognizer(recognizer)
    registry.remove_recognizer("SpacyRecognizer")

    return AnalyzerEngine(
        nlp_engine=nlp_engine,
        registry=registry,
        supported_languages=supported_languages,
        context_aware_enhancer=LemmaContextAwareEnhancer(
            context_similarity_factor=0.35,
            min_score_with_context_similarity=0.4,
        ),
    )



from __future__ import annotations

from enum import Enum

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .base import Scanner

LOGGER = get_logger()

DEFAULT_MODEL = Model(
    path="valurank/distilroberta-bias",
    revision="c1e4a2773522c3acc929a7b2c9af2b7e4137b96d",
    onnx_path="ProtectAI/distilroberta-bias-onnx",
    onnx_revision="3e64d057d20d7ef43fa4f831b992bad28d72640e",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> list[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class Bias(Scanner):
    """
    This class is designed to detect and evaluate potential biases in text using a pretrained model from HuggingFace.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.7,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes the Bias scanner with a probability threshold for bias detection.

        Parameters:
           model (str): The model path to use for bias detection.
           threshold (float): The threshold above which a text is considered biased. Default is 0.7.
           match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
           use_onnx (bool): Whether to use ONNX instead of PyTorch for inference.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._match_type = match_type

        if model is None:
            model = DEFAULT_MODEL

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._classifier = pipeline(
            task="text-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        if output.strip() == "":
            return output, True, 0.0

        highest_score = 0.0
        results_all = self._classifier(self._match_type.get_inputs(prompt + "\n" + output))
        for result in results_all:
            score = round(
                result["score"] if result["label"] == "BIASED" else 1 - result["score"],
                2,
            )

            if score > highest_score:
                highest_score = score

            if score > self._threshold:
                LOGGER.warning(
                    "Detected biased text", highest_score=score, threshold=self._threshold
                )

                return output, False, calculate_risk_score(score, self._threshold)

        LOGGER.debug("Not biased result", highest_score=highest_score, threshold=self._threshold)

        return output, True, 0.0



from __future__ import annotations

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import calculate_risk_score, extract_urls, get_logger

from .base import Scanner

LOGGER = get_logger()
DEFAULT_MODEL = Model(
    path="DunnBC22/codebert-base-Malicious_URLs",
    revision="1221284b2495a4182cdb521be9d755de56e66899",
    onnx_path="ProtectAI/codebert-base-Malicious_URLs-onnx",
    onnx_revision="7bc4fa926eeae5e752d0790cc42faa24eb32fa64",
    pipeline_kwargs={
        "top_k": None,
        "return_token_type_ids": False,
        "max_length": 128,
        "truncation": True,
    },
)

_malicious_labels = [
    "defacement",
    "phishing",
    "malware",
]


class MaliciousURLs(Scanner):
    """
    This scanner is used to scan and detect malicious URLs in the text.

    Large Language Models can occasionally produce malicious URLs in their outputs.
    Using the "DunnBC22/codebert-base-Malicious_URLs" model from HuggingFace, this class classifies
    URLs as either malicious or benign to safeguard users from potential threats.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold=0.5,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes an instance of the MaliciousURLs class.

        Parameters:
            model (Model, optional): The model to use for malicious URL detection.
            threshold (float): The threshold used to determine if the website is malicious. Defaults to 0.5.
            use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
        """

        self._threshold = threshold

        if model is None:
            model = DEFAULT_MODEL

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._classifier = pipeline(
            task="text-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        if output.strip() == "":
            return output, True, 0.0

        urls = extract_urls(output)
        if len(urls) == 0:
            return output, True, 0.0

        LOGGER.debug("Found URLs in the output", len=len(urls))

        highest_malicious_score = 0.0
        results = self._classifier(urls)
        for url, result in zip(urls, results):
            malicious_scores = [
                item["score"] for item in result if item["label"] in _malicious_labels
            ]
            highest_malicious_score = max(malicious_scores)
            if highest_malicious_score > self._threshold:
                LOGGER.warning(
                    "Detected malware URL",
                    url=url,
                    highest_score=highest_malicious_score,
                )

                return output, False, calculate_risk_score(highest_malicious_score, self._threshold)

        LOGGER.debug(
            "Not malware URLs in the output", results=results, highest_score=highest_malicious_score
        )

        return output, True, 0.0



from typing import TypedDict

from llm_guard.model import Model


class NERConfig(TypedDict):
    PRESIDIO_SUPPORTED_ENTITIES: list[str]
    DEFAULT_MODEL: Model
    LABELS_TO_IGNORE: list[str]
    DEFAULT_EXPLANATION: str
    MODEL_TO_PRESIDIO_MAPPING: dict[str, str]
    CHUNK_OVERLAP_SIZE: int
    CHUNK_SIZE: int
    ID_SCORE_MULTIPLIER: float
    ID_ENTITY_NAME: str


BERT_BASE_NER_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
    ],
    "DEFAULT_MODEL": Model(
        path="dslim/bert-base-NER",
        revision="5271995b4b596a059d6efbce30031042aed67cb6",
        onnx_path="dslim/bert-base-NER",
        onnx_revision="5271995b4b596a059d6efbce30031042aed67cb6",
        onnx_subfolder="onnx",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the dslim/bert-base-NER NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "LOC": "LOCATION",
        "ORG": "ORGANIZATION",
        "PER": "PERSON",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}

BERT_LARGE_NER_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
    ],
    "DEFAULT_MODEL": Model(
        path="dslim/bert-large-NER",
        revision="13e784dccceca07aee7a7aab4ad487c605975423",
        onnx_path="dslim/bert-large-NER",
        onnx_revision="13e784dccceca07aee7a7aab4ad487c605975423",
        onnx_subfolder="onnx",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the dslim/bert-large-NER NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "LOC": "LOCATION",
        "ORG": "ORGANIZATION",
        "PER": "PERSON",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}

BERT_ZH_NER_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
    ],
    "DEFAULT_MODEL": Model(
        path="gyr66/bert-base-chinese-finetuned-ner",
        revision="42abc0872240300638223d0e46b9aacdcbcd2906",
        onnx_path="ProtectAI/gyr66-bert-base-chinese-finetuned-ner-onnx",
        onnx_revision="8393f245f01cd153eb2a39fe06a3d4bb37da56fa",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the gyr66/bert-base-chinese-finetuned-ner NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "address": "LOCATION",
        "company": "ORGANIZATION",
        "name": "PERSON",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}

DISTILBERT_AI4PRIVACY_v2_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
        "EMAIL_ADDRESS",
        "PHONE_NUMBER",
        "CREDIT_CARD",
        "CRYPTO",
        "DATE_TIME",
        "IBAN_CODE",
        "IP_ADDRESS",
        "URL",
    ],
    "DEFAULT_MODEL": Model(
        path="Isotonic/distilbert_finetuned_ai4privacy_v2",
        revision="51d7b98bad735ff1af5bda6a6a9a7fbebd102ac9",
        onnx_path="Isotonic/distilbert_finetuned_ai4privacy_v2",
        onnx_revision="51d7b98bad735ff1af5bda6a6a9a7fbebd102ac9",
        onnx_subfolder="onnx",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the Isotonic/distilbert_finetuned_ai4privacy_v2 NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "STREET": "LOCATION",
        "CITY": "LOCATION",
        "ZIPCODE": "LOCATION",
        "BUILDINGNUMBER": "LOCATION",
        "NEARBYGPSCOORDINATES": "LOCATION",
        "SECONDARYADDRESS": "LOCATION",
        "STATE": "LOCATION",
        "COUNTY": "LOCATION",
        "EMAIL": "EMAIL_ADDRESS",
        "COMPANYNAME": "ORGANIZATION",
        "PHONENUMBER": "PHONE_NUMBER",
        "FIRSTNAME": "PERSON",
        "LASTNAME": "PERSON",
        "MIDDLENAME": "PERSON",
        "CREDITCARDNUMBER": "CREDIT_CARD",
        "ETHEREUMADDRESS": "CRYPTO",
        "BITCOINADDRESS": "CRYPTO",
        "LITECOINADDRESS": "CRYPTO",
        "DATE": "DATE_TIME",
        "TIME": "DATE_TIME",
        "DOB": "DATE_OF_BIRTH",
        "IBAN": "IBAN_CODE",
        "IPV4": "IP_ADDRESS",
        "IPV6": "IP_ADDRESS",
        "IP": "IP_ADDRESS",
        "URL": "URL",
        "AGE": "AGE",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}

DEBERTA_AI4PRIVACY_v2_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
        "EMAIL_ADDRESS",
        "PHONE_NUMBER",
        "CREDIT_CARD",
        "CRYPTO",
        "DATE_TIME",
        "IBAN_CODE",
        "IP_ADDRESS",
        "URL",
    ],
    "DEFAULT_MODEL": Model(
        path="Isotonic/deberta-v3-base_finetuned_ai4privacy_v2",
        revision="9ea992753ab2686be4a8f64605ccc7be197ad794",
        onnx_path="Isotonic/deberta-v3-base_finetuned_ai4privacy_v2",
        onnx_revision="9ea992753ab2686be4a8f64605ccc7be197ad794",
        onnx_subfolder="onnx",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the Isotonic/deberta-v3-base_finetuned_ai4privacy_v2 NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "STREET": "LOCATION",
        "CITY": "LOCATION",
        "ZIPCODE": "LOCATION",
        "BUILDINGNUMBER": "LOCATION",
        "NEARBYGPSCOORDINATES": "LOCATION",
        "SECONDARYADDRESS": "LOCATION",
        "STATE": "LOCATION",
        "COUNTY": "LOCATION",
        "EMAIL": "EMAIL_ADDRESS",
        "COMPANYNAME": "ORGANIZATION",
        "PHONENUMBER": "PHONE_NUMBER",
        "FIRSTNAME": "PERSON",
        "LASTNAME": "PERSON",
        "MIDDLENAME": "PERSON",
        "CREDITCARDNUMBER": "CREDIT_CARD",
        "ETHEREUMADDRESS": "CRYPTO",
        "BITCOINADDRESS": "CRYPTO",
        "LITECOINADDRESS": "CRYPTO",
        "DATE": "DATE_TIME",
        "TIME": "DATE_TIME",
        "DOB": "DATE_OF_BIRTH",
        "IBAN": "IBAN_CODE",
        "IPV4": "IP_ADDRESS",
        "IPV6": "IP_ADDRESS",
        "IP": "IP_ADDRESS",
        "URL": "URL",
        "AGE": "AGE",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}

MDEBERTA_AI4PRIVACY_v2_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
        "EMAIL_ADDRESS",
        "PHONE_NUMBER",
        "CREDIT_CARD",
        "CRYPTO",
        "DATE_TIME",
        "IBAN_CODE",
        "IP_ADDRESS",
        "URL",
    ],
    "DEFAULT_MODEL": Model(
        path="Isotonic/mdeberta-v3-base_finetuned_ai4privacy_v2",
        revision="316240516ad48a82ae1b13567670093901e41b5e",
        onnx_path="Isotonic/mdeberta-v3-base_finetuned_ai4privacy_v2",
        onnx_revision="316240516ad48a82ae1b13567670093901e41b5e",
        onnx_subfolder="onnx",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the Isotonic/deberta-v3-base_finetuned_ai4privacy_v2 NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "STREET": "LOCATION",
        "CITY": "LOCATION",
        "ZIPCODE": "LOCATION",
        "BUILDINGNUMBER": "LOCATION",
        "NEARBYGPSCOORDINATES": "LOCATION",
        "SECONDARYADDRESS": "LOCATION",
        "STATE": "LOCATION",
        "COUNTY": "LOCATION",
        "EMAIL": "EMAIL_ADDRESS",
        "COMPANYNAME": "ORGANIZATION",
        "PHONENUMBER": "PHONE_NUMBER",
        "FIRSTNAME": "PERSON",
        "LASTNAME": "PERSON",
        "MIDDLENAME": "PERSON",
        "CREDITCARDNUMBER": "CREDIT_CARD",
        "ETHEREUMADDRESS": "CRYPTO",
        "BITCOINADDRESS": "CRYPTO",
        "LITECOINADDRESS": "CRYPTO",
        "DATE": "DATE_TIME",
        "TIME": "DATE_TIME",
        "DOB": "DATE_OF_BIRTH",
        "IBAN": "IBAN_CODE",
        "IPV4": "IP_ADDRESS",
        "IPV6": "IP_ADDRESS",
        "IP": "IP_ADDRESS",
        "URL": "URL",
        "AGE": "AGE",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}

DEBERTA_LAKSHYAKH93_CONF: NERConfig = {
    "PRESIDIO_SUPPORTED_ENTITIES": [
        "LOCATION",
        "PERSON",
        "ORGANIZATION",
        "EMAIL_ADDRESS",
        "PHONE_NUMBER",
        "CREDIT_CARD",
        "CRYPTO",
        "DATE_TIME",
        "IBAN_CODE",
        "IP_ADDRESS",
        "URL",
    ],
    "DEFAULT_MODEL": Model(
        path="lakshyakh93/deberta_finetuned_pii",
        revision="a038061af92047b0afbbd5ca07d7aa0521789379",
        onnx_path="protectai/lakshyakh93-deberta_finetuned_pii-onnx",
        onnx_revision="af5a07a0cb23bdf0164ef430cc0c965019c9ee24",
        onnx_subfolder="",
        pipeline_kwargs={
            "aggregation_strategy": "simple",
        },
        tokenizer_kwargs={"model_input_names": ["input_ids", "attention_mask"]},
    ),
    "LABELS_TO_IGNORE": ["O", "CARDINAL"],
    "DEFAULT_EXPLANATION": "Identified as {} by the lakshyakh93/deberta_finetuned_pii NER model",
    "MODEL_TO_PRESIDIO_MAPPING": {
        "MISC": "O",
        "BUILDINGNUMBER": "LOCATION",
        "NEARBYGPSCOORDINATE": "LOCATION",
        "STREET": "LOCATION",
        "SECONDARYADDRESS": "LOCATION",
        "PHONE_NUMBER": "PHONE_NUMBER",
        "EMAIL": "EMAIL_ADDRESS",
        "COMPANY_NAME": "ORGANIZATION",
        "FIRSTNAME": "PERSON",
        "FULLNAME": "PERSON",
        "NAME": "PERSON",
        "LASTNAME": "PERSON",
        "MIDDLENAME": "PERSON",
        "DATE": "DATE_TIME",
        "TIME": "DATE_TIME",
        "BITCOINADDRESS": "CRYPTO",
        "URL": "URL",
        "ETHEREUMADDRESS": "CRYPTO",
        "IPV4": "IP_ADDRESS",
        "IPV6": "IP_ADDRESS",
        "CITY": "LOCATION",
        "ZIPCODE": "LOCATION",
        "STREETADDRESS": "LOCATION",
        "CREDITCARDNUMBER": "CREDIT_CARD",
        "STATE": "LOCATION",
        "COUNTY": "LOCATION",
        "SSN": "US_SSN",
        "LITECOINADDRESS": "CRYPTO",
        "IP": "IP_ADDRESS",
        "IBAN": "IBAN_CODE",
    },
    "CHUNK_OVERLAP_SIZE": 40,
    "CHUNK_SIZE": 600,
    "ID_SCORE_MULTIPLIER": 0.4,
    "ID_ENTITY_NAME": "ID",
}



from __future__ import annotations

from enum import Enum

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .ban_substrings import BanSubstrings, MatchType as BanSubstringsMatchType
from .base import Scanner

LOGGER = get_logger()

DEFAULT_MODEL = Model(
    path="ProtectAI/distilroberta-base-rejection-v1",
    revision="65584967c3f22ff7723e5370c65e0e76791e6055",
    onnx_path="ProtectAI/distilroberta-base-rejection-v1",
    onnx_revision="65584967c3f22ff7723e5370c65e0e76791e6055",
    onnx_subfolder="onnx",
    pipeline_kwargs={
        "return_token_type_ids": False,
        "max_length": 128,
        "truncation": True,
    },
)


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> list[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class NoRefusal(Scanner):
    """
    A class used to detect and handle refusals in language model output.

    Refusals are common when the prompt breaches policies defined by the model.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.75,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes an instance of the NoRefusal class.

        Parameters:
            model (Model, optional): The model to use for refusal detection.
            threshold (float): The similarity threshold to consider an output as refusal.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
        """

        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._match_type = match_type

        if model is None:
            model = DEFAULT_MODEL

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._pipeline = pipeline(
            task="text-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        if output.strip() == "":
            return output, True, 0.0

        highest_score = 0.0
        results_all = self._pipeline(self._match_type.get_inputs(output))
        for result in results_all:
            score = round(
                result["score"] if result["label"] == "REJECTION" else 1 - result["score"],
                2,
            )

            if score > highest_score:
                highest_score = score

            if score > self._threshold:
                LOGGER.warning("Detected rejection", highest_score=score)

                return output, False, calculate_risk_score(score, self._threshold)

        LOGGER.debug("No rejection detected", highest_score=highest_score)

        return output, True, 0.0


class NoRefusalLight(BanSubstrings):
    """
    A class used to detect and handle refusals in language model output.

    Refusals are common when the prompt breaches policies defined by the model.
    """

    def __init__(self):
        super().__init__(
            substrings=[
                "I'm sorry",
                "Sorry",
                "I am sorry",
                "I apologize",
                "As an",
                "As an AI",
                "I'm an",
                "I'm just",
                "As a language model",
                "As an Assistant",
                "I cannot",
                "I do not",
                "Hello!",
                "is illegal and unethical",
                "I'm really sorry",
                "OpenAI",
                "unethical",
                "not ethical",
                "not legal",
                "My apologies",
                "but I cannot",
                "If you have any other non-malicious requests",
                "I'm not able to provide",
                "I am really sorry",
                "I cannot create",
                "I cannot support",
                "I can't provide",
                "I can't assist",
            ],
            match_type=BanSubstringsMatchType.STR,
            case_sensitive=False,
            redact=False,
            contains_all=False,
        )



from __future__ import annotations

from enum import Enum

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification, pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .base import Scanner

LOGGER = get_logger()

DEFAULT_MODEL = Model(
    path="papluca/xlm-roberta-base-language-detection",
    revision="9865598389ca9d95637462f743f683b51d75b87b",
    onnx_path="ProtectAI/xlm-roberta-base-language-detection-onnx",
    onnx_revision="dce2fa14a0dc61b6f889537e9ad4fccf083b22bd",
    pipeline_kwargs={
        "top_k": None,
        "return_token_type_ids": False,
        "max_length": 512,
        "truncation": True,
    },
)


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> list[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class Language(Scanner):
    """
    A Scanner subclass responsible for determining the language of a given text
    prompt and verifying its validity against a list of predefined languages.

    Note: when no languages are detected above the threshold, the prompt is considered valid.
    """

    def __init__(
        self,
        valid_languages: list[str],
        *,
        model: Model | None = None,
        threshold: float = 0.6,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes the Language scanner with a list of valid languages.

        Parameters:
            model (Model, optional): A Model object containing the path to the model and its ONNX equivalent.
            valid_languages (Sequence[str]): A list of valid language codes in ISO 639-1.
            threshold (float): Minimum confidence score.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._valid_languages = valid_languages
        self._match_type = match_type

        if model is None:
            model = DEFAULT_MODEL

        tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )

        self._pipeline = pipeline(
            task="text-classification",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **model.pipeline_kwargs,
        )

    def scan(self, prompt: str) -> tuple[str, bool, float]:
        if prompt.strip() == "":
            return prompt, True, 0.0

        results_all = self._pipeline(self._match_type.get_inputs(prompt))
        for result_chunk in results_all:
            languages_above_threshold = [
                result["label"] for result in result_chunk if result["score"] > self._threshold
            ]

            highest_score = max([result["score"] for result in result_chunk])

            # Check if any of the languages above threshold are not valid
            if len(set(languages_above_threshold) - set(self._valid_languages)) > 0:
                LOGGER.warning(
                    "Languages are found with high confidence", languages=languages_above_threshold
                )

                return prompt, False, calculate_risk_score(highest_score, self._threshold)

        LOGGER.debug("Only valid languages are found in the text.")

        return prompt, True, 0.0



from __future__ import annotations

from llm_guard.input_scanners.gibberish import Gibberish as InputGibberish, MatchType
from llm_guard.model import Model

from .base import Scanner


class Gibberish(Scanner):
    """
    A scanner that detects gibberish text.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        threshold: float = 0.7,
        match_type: MatchType | str = MatchType.FULL,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes the Gibberish scanner with a probability threshold for gibberish detection.

        Parameters:
           model (Model, optional): The model used.
           threshold (float): The probability threshold for gibberish detection. Default is 0.7.
           match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
           use_onnx (bool): Whether to use ONNX instead of PyTorch for inference.
        """

        self._scanner = InputGibberish(
            model=model,
            threshold=threshold,
            match_type=match_type,
            use_onnx=use_onnx,
        )

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        return self._scanner.scan(output)



from .analyzer import get_analyzer, get_transformers_recognizer
from .faker import get_fake_value
from .ner_mapping import *
from .regex_patterns import get_regex_patterns

__all__ = [
    "get_analyzer",
    "get_transformers_recognizer",
    "get_fake_value",
    "BERT_BASE_NER_CONF",
    "DEBERTA_AI4PRIVACY_v2_CONF",
    "DISTILBERT_AI4PRIVACY_v2_CONF",
    "BERT_ZH_NER_CONF",
    "BERT_LARGE_NER_CONF",
    "DEBERTA_LAKSHYAKH93_CONF",
    "get_regex_patterns",
]



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
import torch

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer, is_onnx_supported
from llm_guard.util import device, get_logger, lazy_load_dep

from .base import Scanner

LOGGER = get_logger()

MODEL_EN_BGE_BASE = Model(
    path="BAAI/bge-base-en-v1.5",
    revision="a5beb1e3e68b9ab74eb54cfd186867f64f240e1a",
    onnx_path="BAAI/bge-base-en-v1.5",
    onnx_subfolder="onnx",
    onnx_filename="model.onnx",
    onnx_revision="a5beb1e3e68b9ab74eb54cfd186867f64f240e1a",
)
MODEL_EN_BGE_LARGE = Model(
    path="BAAI/bge-large-en-v1.5",
    revision="d4aa6901d3a41ba39fb536a557fa166f842b0e09",
    onnx_path="BAAI/bge-large-en-v1.5",
    onnx_subfolder="onnx",
    onnx_revision="d4aa6901d3a41ba39fb536a557fa166f842b0e09",
)
MODEL_EN_BGE_SMALL = Model(
    path="BAAI/bge-small-en-v1.5",
    revision="5c38ec7c405ec4b44b94cc5a9bb96e735b38267a",
    onnx_path="BAAI/bge-small-en-v1.5",
    onnx_revision="5c38ec7c405ec4b44b94cc5a9bb96e735b38267a",
    onnx_subfolder="onnx",
)


if TYPE_CHECKING:
    import optimum.onnxruntime


class Relevance(Scanner):
    """
    A class used to scan the relevance of the output of a language model to the input prompt.

    This class encodes the prompt and output into vector embeddings, then computes
    the cosine similarity between them. If the similarity is below a given threshold, the output is considered
    not relevant to the prompt.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.5,
        model: Model | None = None,
        use_onnx: bool = False,
    ) -> None:
        """
        Initializes an instance of the Relevance class.

        Parameters:
            threshold: The minimum similarity score to compare prompt and output.
            model: Model for calculating embeddings. Default is `BAAI/bge-base-en-v1.5`.
            use_onnx: Whether to use the ONNX version of the model. Defaults to False.
        """

        self._threshold = threshold

        if model is None:
            model = MODEL_EN_BGE_BASE

        self.pooling_method = "cls"
        self.normalize_embeddings = True

        if use_onnx and is_onnx_supported() is False:
            LOGGER.warning("ONNX is not supported on this machine. Using PyTorch instead of ONNX.")
            use_onnx = False

        if use_onnx:
            optimum_onnxruntime = cast(
                "optimum.onnxruntime",
                lazy_load_dep(
                    "optimum.onnxruntime",
                    "optimum[onnxruntime-gpu]"
                    if device().type == "cuda"
                    else "optimum[onnxruntime]",
                ),
            )
            assert model.onnx_path is not None
            self._model = optimum_onnxruntime.ORTModelForFeatureExtraction.from_pretrained(
                model.onnx_path,
                export=False,
                subfolder=model.onnx_subfolder,
                file_name=model.onnx_filename,
                revision=model.onnx_revision,
                provider=(
                    "CUDAExecutionProvider" if device().type == "cuda" else "CPUExecutionProvider"
                ),
                **model.kwargs,
            )
            LOGGER.debug("Initialized ONNX model", model=model, device=device())
        else:
            transformers = lazy_load_dep("transformers")
            self._model = transformers.AutoModel.from_pretrained(
                model.path, subfolder=model.subfolder, revision=model.revision, **model.kwargs
            ).to(device())
            LOGGER.debug("Initialized model", model=model, device=device())
            self._model.eval()

        self._tokenizer = get_tokenizer(model)

    def pooling(
        self, last_hidden_state: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor | None:
        if self.pooling_method == "cls":
            return last_hidden_state[:, 0]
        elif self.pooling_method == "mean":
            s = torch.sum(last_hidden_state * attention_mask.unsqueeze(-1).float(), dim=1)
            d = attention_mask.sum(dim=1, keepdim=True).float()
            return s / d
        return None

    @torch.no_grad()
    def _encode(self, sentence: str, max_length: int = 512) -> np.ndarray:
        inputs = self._tokenizer(
            [sentence],
            padding=True,
            truncation=True,
            return_tensors="pt",
            max_length=max_length,
        )
        inputs = {key: val.to(device()) for key, val in inputs.items()}

        with torch.no_grad():
            last_hidden_state = self._model(**inputs, return_dict=True).last_hidden_state
            embeddings = self.pooling(last_hidden_state, inputs["attention_mask"])
            assert embeddings is not None
            if self.normalize_embeddings:
                embeddings = torch.nn.functional.normalize(embeddings, dim=-1)

            embeddings = embeddings.cpu().numpy()

        return embeddings[0]

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        if output.strip() == "":
            return output, True, 0.0

        prompt_embedding = self._encode(prompt)
        output_embedding = self._encode(output)
        similarity = prompt_embedding.dot(output_embedding.T)

        if similarity < self._threshold:
            LOGGER.warning("Result is not similar to the prompt", similarity_score=similarity)

            return output, False, round(1 - similarity, 2)

        LOGGER.debug("Result is similar to the prompt", similarity_score=similarity)

        return output, True, 0.0



from __future__ import annotations

from typing import TYPE_CHECKING

from llm_guard.input_scanners.ban_topics import MODEL_DEBERTA_BASE_V2
from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification
from llm_guard.util import device, get_logger, lazy_load_dep

from .base import Scanner

LOGGER = get_logger()

if TYPE_CHECKING:
    import torch

torch = lazy_load_dep("torch")


class FactualConsistency(Scanner):
    """
    FactualConsistency Class:

    This class checks for entailment between a given prompt and output using a pretrained NLI model.
    """

    def __init__(
        self,
        *,
        model: Model | None = None,
        minimum_score=0.75,
        use_onnx=False,
    ) -> None:
        """
        Initializes an instance of the Refutation class.

        Parameters:
            model (Model, optional): The model to use for entailment checking. Defaults to None.
            minimum_score (float): The minimum entailment score for the output to be considered valid. Defaults to 0.75.
            use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
        """

        self._minimum_score = minimum_score

        if model is None:
            model = MODEL_DEBERTA_BASE_V2

        self._tokenizer, self._model = get_tokenizer_and_model_for_classification(
            model=model,
            use_onnx=use_onnx,
        )
        self._model = self._model.to(device())
        if not use_onnx:
            self._model.eval()

    def scan(self, prompt: str, output: str) -> tuple[str, bool, float]:
        if prompt.strip() == "":
            return output, True, 0.0

        tokenized_input_seq_pair = self._tokenizer(
            output, prompt, padding=True, truncation=True, return_tensors="pt"
        )
        tokenized_input_seq_pair = {
            key: val.to(device()) for key, val in tokenized_input_seq_pair.items()
        }

        with torch.no_grad():
            model_output = self._model(**tokenized_input_seq_pair)
            model_prediction = torch.softmax(model_output["logits"][0], -1).tolist()

        label_names = ["entailment", "not_entailment"]
        prediction = {
            name: round(float(pred), 2) for pred, name in zip(model_prediction, label_names)
        }

        entailment_score = prediction["entailment"]
        if entailment_score < self._minimum_score:
            LOGGER.warning("Entailment score is below the threshold", prediction=prediction)

            return output, False, prediction["not_entailment"]

        LOGGER.debug("The output is factually consistent", prediction=prediction)

        return output, True, 0.0



from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, cast

from presidio_analyzer import AnalysisExplanation, EntityRecognizer, RecognizerResult
from presidio_analyzer.nlp_engine import NlpArtifacts
from transformers import TokenClassificationPipeline

from llm_guard.model import Model
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_ner
from llm_guard.util import get_logger, lazy_load_dep, split_text_to_word_chunks

from .ner_mapping import BERT_BASE_NER_CONF

LOGGER = get_logger()

if TYPE_CHECKING:
    import transformers


class TransformersRecognizer(EntityRecognizer):
    """
    Wrapper for a transformers_rec model, if needed to be used within Presidio Analyzer.
    The class loads models hosted on HuggingFace - https://huggingface.co/
    and loads the model and tokenizer into a TokenClassification pipeline.
    Samples are split into short text chunks, ideally shorter than max_length input_ids of the individual model,
    to avoid truncation by the Tokenizer and loss of information

    A configuration object should be maintained for each dataset-model combination and translate
    entities names into a standardized view. A sample of a configuration file is attached in
    the example.
    :param supported_entities: List of entities to run inference on
    :type supported_entities: Optional[List[str]]
    :param pipeline: Instance of a TokenClassificationPipeline including a Tokenizer and a Model, defaults to None
    :type pipeline: Optional[TokenClassificationPipeline], optional
    :param model_path: string referencing a HuggingFace uploaded model to be used for Inference, defaults to None
    :type model_path: Optional[str], optional

    :example
    >from presidio_analyzer import AnalyzerEngine, RecognizerRegistry
    >model_path = "obi/deid_roberta_i2b2"
    >transformers_recognizer = TransformersRecognizer(model_path=model_path,
    >supported_entities = model_configuration.get("PRESIDIO_SUPPORTED_ENTITIES"))
    >transformers_recognizer.load_transformer(**model_configuration)
    >registry = RecognizerRegistry()
    >registry.add_recognizer(transformers_recognizer)
    >analyzer = AnalyzerEngine(registry=registry)
    >sample = "My name is Christopher and I live in Irbid."
    >results = analyzer.analyze(sample, language="en",return_decision_process=True)

    >for result in results:
    >    print(result,'----', sample[result.start:result.end])
    """

    ignore_labels: list[str]
    model_to_presidio_mapping: dict[str, str]
    default_explanation: str
    text_overlap_length: int
    chunk_length: int
    id_entity_name: str
    id_score_reduction: float
    pipeline: transformers.Pipeline | None

    def load(self) -> None:
        pass

    def __init__(
        self,
        model: Model,
        pipeline: TokenClassificationPipeline | None = None,
        supported_entities: list[str] | None = None,
        supported_language: str = "en",
    ):
        if not supported_entities:
            supported_entities = BERT_BASE_NER_CONF["PRESIDIO_SUPPORTED_ENTITIES"]
        super().__init__(
            supported_entities=supported_entities,
            name=f"Transformers model {model.path}",
        )

        self.model = model
        self.pipeline = pipeline
        self.is_loaded = False
        self.supported_language = supported_language

    def load_transformer(
        self,
        use_onnx: bool = False,
        **kwargs,
    ) -> None:
        """Load external configuration parameters and set default values.

        :param use_onnx: flag to use ONNX optimized model
        :type use_onnx: bool, optional
        :param kwargs: define default values for class attributes and modify pipeline behavior
        **MODEL_TO_PRESIDIO_MAPPING (dict) -  defines mapping entity strings from chosen model format to Presidio format
        **CHUNK_OVERLAP_SIZE (int) - number of overlapping characters in each text chunk
        when splitting a single text into multiple inferences
        **CHUNK_SIZE (int) - number of characters in each chunk of text
        **LABELS_TO_IGNORE (List(str)) - List of entities to skip evaluation. Defaults to ["O"]
        **DEFAULT_EXPLANATION (str) - string format to use for prediction explanations
        **ID_ENTITY_NAME (str) - name of the ID entity
        **ID_SCORE_REDUCTION (float) - score multiplier for ID entities
        **use_onnx (bool) - flag to use ONNX optimized model
        """

        self.model_to_presidio_mapping = kwargs.get("MODEL_TO_PRESIDIO_MAPPING", {})
        self.ignore_labels = kwargs.get("LABELS_TO_IGNORE", ["O"])
        self.default_explanation = kwargs.get("DEFAULT_EXPLANATION", "")
        self.text_overlap_length = kwargs.get("CHUNK_OVERLAP_SIZE", 40)
        self.chunk_length = kwargs.get("CHUNK_SIZE", 600)
        self.id_entity_name = kwargs.get("ID_ENTITY_NAME", "ID")
        self.id_score_reduction = kwargs.get("ID_SCORE_REDUCTION", 0.5)

        if not self.pipeline:
            if not self.model:
                self.model = Model(
                    path="dslim/bert-base-NER",
                    onnx_path="dslim/bert-base-NER",
                    onnx_subfolder="onnx",
                )
                LOGGER.warning(
                    "'model' argument is None. Using default",
                    model=self.model,
                )

            self._load_pipeline(
                use_onnx=use_onnx,
            )

    def _load_pipeline(
        self,
        use_onnx: bool = False,
    ) -> None:
        """Initialize NER transformers_rec pipeline using the model_path provided"""
        tf_tokenizer, tf_model = get_tokenizer_and_model_for_ner(self.model, use_onnx=use_onnx)

        self.model.pipeline_kwargs["ignore_labels"] = self.ignore_labels

        transformers = cast("transformers", lazy_load_dep("transformers"))
        self.pipeline = transformers.pipeline(
            "ner",
            model=tf_model,
            tokenizer=tf_tokenizer,
            **self.model.pipeline_kwargs,
        )

        self.is_loaded = True

    def get_supported_entities(self) -> list[str]:
        """
        Return supported entities by this model.
        :return: List of the supported entities.
        """
        return self.supported_entities

    # Class to use transformers_rec with Presidio as an external recognizer.
    def analyze(
        self, text: str, entities: list[str], nlp_artifacts: NlpArtifacts | None = None
    ) -> list[RecognizerResult]:
        """
        Analyze text using transformers_rec model to produce NER tagging.
        :param text : The text for analysis.
        :param entities: Not working properly for this recognizer.
        :param nlp_artifacts: Not used by this recognizer.
        :return: The list of Presidio RecognizerResult constructed from the recognized
            transformers_rec detections.
        """

        results = list()
        # Run transformer model on the provided text
        ner_results = self._get_ner_results_for_text(text)

        for res in ner_results:
            res["entity_group"] = self.__check_label_transformer(res["entity_group"])
            if not res["entity_group"] or res["entity_group"] not in entities:
                continue

            if res["entity_group"] == self.id_entity_name:
                LOGGER.debug(
                    "ID entity found, multiplying score", score_reduction=self.id_score_reduction
                )
                res["score"] = res["score"] * self.id_score_reduction

            textual_explanation = self.default_explanation.format(res["entity_group"])
            explanation = self.build_transformers_explanation(
                float(round(res["score"], 2)), textual_explanation, res["word"]
            )

            word = text[res["start"] : res["end"]]
            if word[0] == " ":
                res["start"] += 1
            transformers_result = self._convert_to_recognizer_result(res, explanation)

            results.append(transformers_result)

        return results

    def _get_ner_results_for_text(self, text: str) -> list[dict]:
        """The function runs model inference on the provided text.
        The text is split into chunks with n overlapping characters.
        The results are then aggregated and duplications are removed.

        :param text: The text to run inference on
        :type text: str
        :return: List of entity predictions on the word level
        :rtype: List[dict]
        """
        assert self.pipeline is not None
        assert self.pipeline.tokenizer is not None

        model_max_length = self.pipeline.tokenizer.model_max_length
        # calculate inputs based on the text
        text_length = len(text)
        # split text into chunks
        if text_length <= model_max_length:
            predictions = self.pipeline(text)  # type: ignore
        else:
            LOGGER.info(
                "splitting the text into chunks",
                length=text_length,
                model_max_length=model_max_length,
            )
            predictions: list[dict[str, Any]] = []
            chunk_indexes = split_text_to_word_chunks(
                text_length, self.chunk_length, self.text_overlap_length
            )

            # iterate over text chunks and run inference
            for chunk in chunk_indexes:
                chunk_text = text[chunk.start : chunk.end]
                chunk_preds = self.pipeline(chunk_text)

                assert isinstance(chunk_preds, list)

                # align indexes to match the original text - add to each position the value of chunk_start
                aligned_predictions: list[dict[str, Any]] = []
                for prediction in chunk_preds:
                    prediction_tmp: dict[str, Any] = copy.deepcopy(prediction)
                    prediction_tmp["start"] += chunk.start
                    prediction_tmp["end"] += chunk.start
                    aligned_predictions.append(prediction_tmp)

                predictions.extend(aligned_predictions)

        # remove duplicates
        predictions = [dict(t) for t in {tuple(d.items()) for d in predictions}]
        return predictions

    @staticmethod
    def _convert_to_recognizer_result(
        prediction_result: dict, explanation: AnalysisExplanation
    ) -> RecognizerResult:
        """The method parses NER model predictions into a RecognizerResult format to enable down the stream analysis

        :param prediction_result: A single example of entity prediction
        :type prediction_result: dict
        :param explanation: Textual representation of model prediction
        :type explanation: str
        :return: An instance of RecognizerResult which is used to model evaluation calculations
        :rtype: RecognizerResult
        """

        transformers_results = RecognizerResult(
            entity_type=prediction_result["entity_group"],
            start=prediction_result["start"],
            end=prediction_result["end"],
            score=float(round(prediction_result["score"], 2)),
            analysis_explanation=explanation,
        )

        return transformers_results

    def build_transformers_explanation(
        self,
        original_score: float,
        explanation: str,
        pattern: str,
    ) -> AnalysisExplanation:
        """
        Create explanation for why this result was detected.
        :param original_score: Score given by this recognizer
        :param explanation: Explanation string
        :param pattern: Regex pattern used
        :return Structured explanation and scores of a NER model prediction
        :rtype: AnalysisExplanation
        """
        return AnalysisExplanation(
            recognizer=self.__class__.__name__,
            original_score=float(original_score),
            textual_explanation=explanation,
            pattern=pattern,
        )

    def __check_label_transformer(self, label: str) -> str | None:
        """The function validates the predicted label is identified by Presidio
        and maps the string into a Presidio representation
        :param label: Predicted label by the model
        :return: Returns the adjusted entity name
        """

        # convert model label to presidio label
        entity = self.model_to_presidio_mapping.get(label, None)

        if entity in self.ignore_labels:
            return None

        if entity is None:
            LOGGER.warning("Found unrecognized label, returning entity as is", label=label)
            return label

        if entity not in self.supported_entities:
            LOGGER.warning("Found entity which is not supported by Presidio", entity=entity)
            return entity
        return entity
